# coding: utf-8

from .project_name_window import check_project_name
from .save_popup import Save_popup
from .settings_window import Settings_window
from .splash_window import Splash_window
from .structure_classes import Settings, Nucleus, Fiber, Nuclei, Fibers, \
    Graphical_element, Selection_box, Table_items, Table_entry
from .warning_window import Warning_window
from ._check_image import check_image



# coding: utf-8

from tkinter import Toplevel, ttk, IntVar
from screeninfo import get_monitors


class Warning_window(Toplevel):
    """Popup window warning the user that the selected action may cause a data
    loss, and proposing to user save before continuing, continue without
    saving, or cancel."""

    def __init__(self, main_window, return_var: IntVar) -> None:
        """Creates the window, arranges its layout and displays it.

        Args:
            main_window: The main window of the project.
            return_var: A variable indicating the option the user chose.
        """

        super().__init__(main_window)

        # Setting variable
        self._return_var = return_var

        # Setting window properties
        self.resizable(False, False)
        self.grab_set()
        self.title("Hold on!")
        self.bind('<Destroy>', self._cancel)

        # Setting the layout and displaying the window
        self._set_layout()
        self.update()
        self._center()

    def _set_layout(self) -> None:
        """Creates the labels and the buttons, places them and displays
        them."""

        # Setting the label
        ttk.Label(self,
                  text="Are you sure about closing an unsaved project ?"). \
            pack(anchor='n', expand=False, fill='none', side='top',
                 padx=20, pady=20)

        # Setting the buttons
        ttk.Button(self, text='Save Before Continuing',
                   command=self._save, width=40). \
            pack(anchor='n', expand=False, fill='none', side='top',
                 padx=20, pady=7)
        ttk.Button(self, text='Continue Without Saving',
                   command=self._ignore,
                   width=40). \
            pack(anchor='n', expand=False, fill='none', side='top',
                 padx=20, pady=7)
        ttk.Button(self, text='Cancel', command=self.destroy, width=40). \
            pack(anchor='n', expand=False, fill='none', side='top',
                 padx=20, pady=7)

    def _center(self) -> None:
        """Centers the popup window on the currently used monitor."""

        # Getting the current monitor
        monitors = get_monitors()
        candidates = [monitor for monitor in monitors if
                      0 <= self.winfo_x() - monitor.x <= monitor.width and
                      0 <= self.winfo_y() - monitor.y <= monitor.height]
        current_monitor = candidates[0] if candidates else monitors[0]

        # Getting the parameters of interest
        scr_width = current_monitor.width
        scr_height = current_monitor.height
        x_offset = current_monitor.x
        y_offset = current_monitor.y
        height = self.winfo_height()
        width = self.winfo_width()

        # Actually centering the window
        self.geometry('+%d+%d' % (x_offset + (scr_width - width) / 2,
                                  y_offset + (scr_height - height) / 2))

    def _save(self) -> None:
        """Simply sets the return flag to 2 in case the user wants to save
        before continuing."""

        self._return_var.set(2)

    def _ignore(self) -> None:
        """Simply sets the return flag to 1 in case the user wants to continue
        without saving."""

        self._return_var.set(1)

    def _cancel(self, *_, **__) -> None:
        """Simply sets the return flag to 0 in case the user cancels or closes
        the window."""

        self._return_var.set(0)



# coding: utf-8

from tkinter import ttk, filedialog, Tk, IntVar, PhotoImage, Menu, StringVar, \
    messagebox, Checkbutton, BooleanVar
from platform import system, release
from shutil import rmtree
from webbrowser import open_new
from threading import Thread, Event
from queue import Empty, Queue
from time import sleep
from pickle import load, dump
from functools import partial, wraps
from pathlib import Path
from typing import Callable, Optional
import logging
from pkg_resources import resource_filename

from .tools import Settings
from .tools import Save_popup
from .tools import Warning_window
from .tools import Settings_window
from .tools import Splash_window
from .tools import check_project_name
from .files_table import Files_table
from .image_canvas import Image_canvas

# Sets a better resolution on recent Windows platforms
if system() == "Windows" and int(release()) >= 8:
    from ctypes import windll
    windll.shcore.SetProcessDpiAwareness(True)

# Todo:
#   Set up unit tests


def _save_before_closing(func: Callable) -> Callable:
    """Decorator for warning the user when an action that may cause unwanted
    data loss is triggered.

    The user will then be proposed to save the work, continue anyway, or abort.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """The warning window will return False if the operation shouldn't be
        carried on"""

        if self._create_warning_window():
            func(self, *args, **kwargs)
    return wrapper


class Main_window(Tk):
    """The main window of the MyoFInDer software.

    It manages all the buttons, menus, events, and the secondary windows.
    """

    def __init__(self, app_folder: Optional[Path]) -> None:
        """Creates the splash window, then the main window, sets the layout and
        the callbacks.

        Args:
            app_folder: The Path to the application folder.
        """

        # Setting the logger
        self._logger = logging.getLogger("MyoFInDer.MainWindow")

        # Generates a splash window while waiting for the modules to load
        self.log("Creating the splash window")
        splash = Splash_window()
        self.log("Centering the splash window")
        splash.resize_image()
        self.log("Displaying the splash window")
        self._segmentation = splash.display()
        splash.destroy()

        # Initializes the main window
        super().__init__()
        self.title("MyoFInDer - New Project (Unsaved)")
        if system() == "Windows":
            self.state('zoomed')
        else:
            self.attributes('-zoomed', True)
        self.app_folder = app_folder

        # Sets the application icon
        self.log("Setting the application icon")
        icon = PhotoImage(file=resource_filename(
            'myofinder', 'app_images/project_icon.png'))
        self.iconphoto(False, icon)  # Without it the icon is buggy in Windows
        self.iconphoto(True, icon)

        # Sets the settings, variables, callbacks, menus and layout
        self._set_variables()
        self._set_traces()
        self._set_menu()
        self._set_layout()

        # Sets the image canvas and the files table
        self.log("Creating the image canvas")
        self._image_canvas = Image_canvas(self._frm, self)
        self.log("Creating the files table")
        self._files_table = Files_table(self._aux_frame, self)
        self._image_canvas.nuclei_table = self._files_table
        self._files_table.image_canvas = self._image_canvas

        # Loading the previous setting if applicable
        if self.app_folder is not None:
            if (self.app_folder / 'settings.pickle').exists():
                self._load_settings(self.app_folder)

        # Finishes the initialization and starts the event loop
        self.update()
        self.log("Setting the main windows's bindings and protocols")
        self.bind_all('<Control-s>', self._save_button_pressed)
        self.protocol("WM_DELETE_WINDOW", self._safe_destroy)

    def set_unsaved_status(self) -> None:
        """Sets the title and save button when a project is modified."""

        self.log("Setting the unsaved status")

        if self._current_project is not None:
            self.title(f"MyoFInDer - Project "
                       f"'{self._current_project.name}' (Unsaved)")
            self._save_button['state'] = 'enabled'
            self._save_button['text'] = 'Save'
        else:
            self.title("MyoFInDer - New Project (Unsaved)")
            self._save_button['state'] = 'enabled'
            self._save_button['text'] = 'Save As'

    def save_settings(self, project_path: Path) -> None:
        """Saves the settings to a settings.pickle file.

        Args:
            project_path: The Path where the project will be saved.
        """

        # Saving the settings
        settings = {key: value.get() for key, value
                    in vars(self.settings).items()}
        self.log(f"Settings values: {str(self.settings)}")

        settings_file = project_path / 'settings.pickle'
        with open(settings_file, 'wb+') as param_file:
            dump(settings, param_file, protocol=4)
            self.log(f"Saved the settings at: {settings_file}")

    def update_master_check(self) -> None:
        """Updates the master checkbox according to the states of the
        checkboxes of all the files menu entries."""

        self._master_check_var.set(self._files_table.all_checked)

    def log(self, msg: str) -> None:
        """Wrapper for reducing the verbosity of logging."""

        # For some reason the processing adds a handler to the root logger
        root = self._logger.parent.parent
        while root.hasHandlers():
            root.removeHandler(root.handlers[0])

        self._logger.log(logging.INFO, msg)

    def _set_variables(self) -> None:
        """Sets the different variables used in the class."""

        self.log("Setting the main windows's variables")

        self.settings = Settings()

        # Variables used when there's a conflict in the choice of colors for
        # the fibers and nuclei
        self._previous_nuclei_colour = StringVar(
            value=self.settings.nuclei_colour.get())
        self._previous_fiber_colour = StringVar(
            value=self.settings.fiber_colour.get())

        # The variables associated with the master checkbox
        self._master_check_var = BooleanVar(value=True)

        # Variables managing the processing thread
        self._processed_images_count = IntVar(value=0)
        self._img_to_process_count = 0
        self._stop_thread = False
        self._stop_event = Event()
        self._stop_event.set()
        self._queue = Queue(maxsize=0)
        self._thread = Thread(target=self._process_thread)
        self._thread.start()

        self._current_project: Optional[Path] = None  # Path to current project

    def _set_traces(self) -> None:
        """Sets the callbacks triggered upon modification of the settings."""

        self.log("Setting the main windows's traces")

        # Making sure there's no conflict between the nuclei and fibers colors
        self.settings.fiber_colour.trace_add("write", self._nuclei_colour_sel)
        self.settings.nuclei_colour.trace_add("write", self._nuclei_colour_sel)

        # Some settings should enable the save button when modified
        self.settings.save_overlay.trace_add(
            "write", self._enable_save_button)
        self.settings.fiber_threshold.trace_add("write",
                                                self._enable_save_button)
        self.settings.small_objects_threshold.trace_add(
            "write", self._enable_save_button)

        # Updates the display when an image has been processed
        self._processed_images_count.trace_add("write",
                                               self._update_processed_images)

    def _set_menu(self) -> None:
        """Sets the menu bar."""

        self.log("Setting the main windows's menu bar")

        # Sets the overall menu bar
        self._menu_bar = Menu(self)
        self.config(menu=self._menu_bar)
        self._menu_bar.config(font=("TKDefaultFont", 12))

        # Sets the file menu
        self._file_menu = Menu(self._menu_bar, tearoff=0)
        self._file_menu.add_command(label="New Empty Project",
                                    command=self._safe_empty_project)
        self._file_menu.add_command(label="Save Project As",
                                    command=partial(self._save_button_pressed,
                                                    force_save_as=True))
        self._file_menu.add_command(label="Delete Current Project",
                                    command=self._delete_current_project)
        self._file_menu.entryconfig(
            self._file_menu.index("Delete Current Project"), state="disabled")
        self._file_menu.add_command(
            label="Load From Explorer",
            command=self._safe_load)

        self._menu_bar.add_cascade(label="File", menu=self._file_menu)

        # Sets the settings menu
        self._settings_menu = Menu(self._menu_bar, tearoff=0)
        self._settings_menu.add_command(
            label="Settings", command=partial(Settings_window, self))
        self._menu_bar.add_cascade(label="Settings", menu=self._settings_menu)

        # Sets the help menu
        self._help_menu = Menu(self._menu_bar, tearoff=0)
        self._help_menu.add_command(label="Help", command=self._open_github)
        self._menu_bar.add_cascade(label="Help", menu=self._help_menu)

        # Sets the quit menu
        self._quit_menu = Menu(self._menu_bar, tearoff=0)
        self._quit_menu.add_command(label="Quit", command=self._safe_destroy)
        self._menu_bar.add_cascade(label="Quit", menu=self._quit_menu)

    def _set_layout(self) -> None:
        """Sets the overall layout of the window."""

        self.log("Setting the main windows's layout")

        # The main frame of the window
        self._frm = ttk.Frame()
        self._frm.pack(fill='both', expand=True)

        # The frame containing the files table, the buttons and the checkboxes
        self._aux_frame = ttk.Frame(self._frm)
        self._aux_frame.pack(expand=False, fill="both",
                             anchor="e", side='right')

        # The row of buttons
        self._button_frame = ttk.Frame(self._aux_frame)
        self._button_frame.pack(expand=False, fill="x", anchor='n', side='top')

        # The first row of checkboxes
        self._tick_frame_1 = ttk.Frame(self._aux_frame)
        self._tick_frame_1.pack(expand=False, fill="x", anchor='n', side='top')

        # The second row of checkboxes
        self._tick_frame_2 = ttk.Frame(self._aux_frame)
        self._tick_frame_2.pack(expand=False, fill="x", anchor='n', side='top')

        self._text_button_frame = ttk.Frame(self._aux_frame)
        self._text_button_frame.pack(anchor="n", side="top", fill='x', padx=3,
                                     pady=5)

        # Creating the buttons
        self._load_button = ttk.Button(self._button_frame, text="Load Images",
                                       command=self._select_images)
        self._load_button.pack(fill="x", anchor="w", side='left', padx=10,
                               pady=5, expand=True)
        self._button_style = ttk.Style()
        self._process_images_button = ttk.Button(self._button_frame,
                                                 text="Process Images",
                                                 command=self._process_images,
                                                 state='disabled')
        self._process_images_button.pack(fill="x", anchor="w", side='left',
                                         padx=10, pady=5, expand=True)
        self._save_button = ttk.Button(self._button_frame, text='Save As',
                                       command=self._save_button_pressed,
                                       state='enabled')
        self._save_button.pack(fill="x", anchor="w", side='left', padx=10,
                               pady=5, expand=True)

        # Creating the checkboxes of the first row
        self._channels = ttk.Label(self._tick_frame_1, text="  Channels :   ")
        self._channels.pack(anchor="w", side="left", fill='x', padx=3, pady=5)
        self._blue_channel_check_button = ttk.Checkbutton(
            self._tick_frame_1, text="Blue", onvalue=True, offvalue=False,
            variable=self.settings.blue_channel_bool,
            command=self._set_image_channels)
        self._blue_channel_check_button.pack(anchor="w", side="left", fill='x',
                                             padx=5, pady=5)
        self._green_channel_check_button = ttk.Checkbutton(
            self._tick_frame_1, text="Green", onvalue=True, offvalue=False,
            variable=self.settings.green_channel_bool,
            command=self._set_image_channels)
        self._green_channel_check_button.pack(anchor="w", side="left",
                                              fill='x', padx=5, pady=5)
        self._red_channel_check_button = ttk.Checkbutton(
            self._tick_frame_1, text="Red", onvalue=True, offvalue=False,
            variable=self.settings.red_channel_bool,
            command=self._set_image_channels)
        self._red_channel_check_button.pack(anchor="w", side="left", fill='x',
                                            padx=5, pady=5)

        # Creating the checkboxes of the second row
        self._indicator = ttk.Label(self._tick_frame_2, text="  Indicators : ")
        self._indicator.pack(anchor="w", side="left", fill='x', padx=3, pady=5)
        self._show_nuclei_check_button = ttk.Checkbutton(
            self._tick_frame_2, text="Nuclei", onvalue=True, offvalue=False,
            variable=self.settings.show_nuclei, command=self._set_indicators)
        self._show_nuclei_check_button.pack(anchor="w", side="left", fill='x',
                                            padx=3, pady=5)
        self._show_fibers_check_button = ttk.Checkbutton(
            self._tick_frame_2, text="Fibers", onvalue=True, offvalue=False,
            variable=self.settings.show_fibers, command=self._set_indicators)
        self._show_fibers_check_button.pack(anchor="w", side="left", fill='x',
                                            padx=3, pady=5)

        # Label displaying info during image processing
        self._processing_label = ttk.Label(self._text_button_frame, text="")
        self._processing_label.pack(anchor="w", side="left", fill='x',
                                    padx=(9, 0))

        self._delete_all_button = ttk.Button(self._text_button_frame,
                                             text='X', width=2,
                                             command=self._delete_many_files)
        self._delete_all_button.pack(anchor="e", side="right", fill='none',
                                     padx=(0, 15))

        self._check_img = PhotoImage(width=1, height=1)
        self._select_all_button = Checkbutton(self._text_button_frame,
                                              image=self._check_img,
                                              width=6, height=24,
                                              command=self._invert_checkboxes,
                                              variable=self._master_check_var)
        self._select_all_button.pack(anchor="e", side="right", fill='none')

    def _enable_save_button(self, _, __, ___) -> None:
        """Some settings should enable the save button when modified."""

        self._save_button['state'] = 'enabled'

    def _load_settings(self, project_path: Path) -> None:
        """Loads the settings from the settings file if any is present.

        Args:
            project_path: The Path where the project is saved.
        """

        # Gets the settings from the settings.pickle file
        settings_file = project_path / 'settings.pickle'
        if settings_file.is_file() and settings_file.exists():
            self.log(f"Loading the settings file from {settings_file}")
            with open(settings_file, 'rb') as param_file:
                settings = load(param_file)
        else:
            self.log("No setting file detected, loading the default settings")
            settings = dict()

        # Creates a Settings instance and sets the values
        self.settings.update(settings)
        self.log(f"Settings values: {str(self.settings)}")

    def _delete_current_project(self) -> None:
        """Deletes the current project and all the associated files."""

        self.log("User requested deletion of the current project")

        if self._current_project is None:
            self.log("No current project to delete")
            return

        # Security to prevent unwanted data loss
        ret = messagebox.askyesno('Hold on !',
                                  "Do you really want to delete the current "
                                  "project ?\nThis operation can't be undone.")

        if ret:
            self.log("User approved deletion of current project")

            # Simply deletes all the project files
            rmtree(self._current_project)
            self.log("Deleted the files of the current project")

            # Creates a new empty project
            self._create_empty_project()

        else:
            self.log("User aborted deletion of current project")

    def _create_empty_project(self) -> None:
        """Creates a new empty project."""

        self.log("Creating an empty project")

        # Resets the entire window
        self._files_table.table_items.reset()
        self._image_canvas.reset()

        # Sets the buttons and the menu
        self._save_button['state'] = 'enabled'
        self._save_button['text'] = 'Save As'
        self._process_images_button['state'] = 'disabled'
        self._file_menu.entryconfig(self._file_menu.index("Delete Current "
                                                          "Project"),
                                    state="disabled")

        # Sets the title
        self.title("MyoFInDer - New Project (Unsaved)")
        self._current_project = None

    def _save_project(self, directory: Path) -> None:
        """Saves a project, its images and the associated data.

        Args:
            directory: The path to the folder where the project should be
                saved.
        """

        # Setting the save button, the project title and the menu entries
        self._set_title_and_button(directory)

        # Creating the folder if needed
        if not (directory.is_dir() and directory.exists()):
            self.log(f"Creating the directory where to save the "
                     f"project at {directory}")
            Path.mkdir(directory, parents=True)

        # Displays a popup indicating the project is being saved
        saving_popup = Save_popup(self, directory)
        sleep(1)

        # Actually saving the project
        self.save_settings(directory)
        save_overlay = self.settings.save_overlay.get()
        self._files_table.save_project(directory, save_overlay)
        self.log(f"Project saved at {directory}")

        saving_popup.destroy()

    def _load_project(self) -> None:
        """Loads a project, its images and its data."""

        self.log("Project loading requested by the user")

        # Choose the folder in a dialog window
        directory = filedialog.askdirectory(
            initialdir=Path.cwd(),
            mustexist=True,
            title="Choose a Project Folder")

        if not directory:
            self.log("Project loading aborted bu the user")
            return

        directory = Path(directory)
        self.log(f"User requested to load project {directory}")

        # Checking that a valid project was selected
        if not (directory.is_dir() and directory.exists() and
                (directory / 'data.pickle').is_file() and
                (directory / 'data.pickle').exists()):
            messagebox.showerror("Error while loading",
                                 "This isn't a valid MyoFInDer project !")
            self.log("The selected directory is not valid for loading into "
                     "MyoFInDer")
            return

        # Setting the save button, the project title and the menu entries
        self._set_title_and_button(directory)

        # Actually loading the project
        self.log(f"Loading the project {directory}")
        self._load_settings(directory)
        self._files_table.load_project(directory)

        # Enabling the process images button
        if self._files_table.table_items:
            self._process_images_button['state'] = 'enabled'
        else:
            self._process_images_button['state'] = 'disabled'

    def _set_title_and_button(self, directory: Path) -> None:
        """Sets the project title and the save button when loading or saving
        a project.

        Args:
            directory: The path to the directory where the project is saved.
        """

        self.log("Setting the title and buttons of the main window")

        # Sets the save button
        self._save_button['state'] = 'disabled'
        self._save_button['text'] = 'Save'

        # Sets the menu
        self._file_menu.entryconfig(
            self._file_menu.index("Delete Current Project"), state="normal")

        # Sets the project name
        self._current_project = directory
        self.title(f"MyoFInDer - Project '{directory.name}'")

    def _create_warning_window(self) -> bool:
        """Creates a warning window in case the user triggers an action that
        might cause unwanted data loss.

        Also manages the situation when the user tries to exit the program
        while images are being processed.
        """

        self.log("Checking if it is necessary to create a warning window")

        # Case when the user tries to exit while computations are running
        if not self._queue.empty():
            self.log("Processing queue not empty")
            ret = messagebox.askyesno(
                'Hold on !',
                "The program is still computing !\n"
                "It is safer to wait for the computation to end.\n"
                "Quit anyway ?\n"
                "(resources may take some time to be released even after "
                "exiting)")

            # Trying to stop the computation if requested
            if ret:
                self._stop_thread = True
                self.log("User requested the computation to stop")
            else:
                self.log("User canceled the action")

            return ret

        # If the project is unsaved, propose to save it
        if self._save_button['state'] == 'enabled' and \
                self._files_table.table_items:

            self.log("The project is unsaved")

            # Creating the warning window and waiting for the user to choose
            return_var = IntVar()
            warning_window = Warning_window(self, return_var)
            self.wait_variable(return_var)

            # Handling the different possible answers from the user
            ret = return_var.get()
            # The user wants to save and proceed
            if ret == 2:
                self.log("User requested to save the project and proceed")
                warning_window.destroy()
                return self._save_button_pressed()
            # the user wants to proceed without saving
            elif ret == 1:
                warning_window.destroy()
                self.log("User requested to proceed anyway without saving")
                return True
            # The user cancelled its action
            else:
                self.log("User requested to cancel the action")
                return False

        self.log("No warning window required")
        return True

    def _save_button_pressed(self,
                             _: Optional[Event] = None,
                             force_save_as: bool = False) -> bool:
        """Method called when a save action is triggered by the user or when
        CTRL+S is hit.

        It may or may not lead to an actual save.

        Args:
            force_save_as: If True, saves the project as a new one even if it
                already has a folder.
        """

        self.log("The user requested to save the current project")

        # Asks for a new project name if needed
        if force_save_as or self._current_project is None:
            save = False
            path = None

            # Letting the user choose a path for saving the project
            while not save:
                # Ask for a project name and location
                path = filedialog.asksaveasfilename(
                    confirmoverwrite=False, defaultextension=None,
                    initialdir=Path.cwd(), title='Saving Current Project')
                # The user can exit by pressing the cancel button
                if not path:
                    self.log("The user aborted the save action")
                    return False
                # Checking the validity of the chosen path
                path = Path(path)
                save = check_project_name(path)

            self.log(f"Saving the project to {path}")
            self._save_project(path)
            return True

        # Perform a normal save otherwise
        else:
            self.log(f"Saving the project to {self._current_project}")
            self._save_project(self._current_project)
            return True

    @_save_before_closing
    def _safe_destroy(self) -> None:
        """Stops the computation thread, closes the main window, and displays
        a warning if there's unsaved data."""

        self.log("Requesting the processing thread to finish")
        self._stop_thread = True
        sleep(0.5)
        self.log("Waiting for the processing thread to finish")
        self._thread.join(timeout=1)

        if self._thread.is_alive():
            self.log("ERROR! The processing thread is still alive !")
        else:
            self.log("The processing thread terminated as excepted")

        self.log("Destroying the main window")
        self.destroy()

    @_save_before_closing
    def _safe_empty_project(self) -> None:
        """Creates an empy project, and displays a warning if there's unsaved
        data."""

        self._create_empty_project()

    @_save_before_closing
    def _safe_load(self) -> None:
        """Loads an existing project, and displays a warning if there's unsaved
        data."""

        self._load_project()

    def _disable_buttons(self) -> None:
        """Disables most of the buttons and menu entries when computing."""

        self.log("Disabling the buttons")

        # Disabling the buttons
        self._load_button['state'] = "disabled"
        self._save_button['state'] = "disabled"
        self._select_all_button['state'] = 'disabled'
        self._delete_all_button['state'] = 'disabled'

        # Disabling the menu entries
        self._file_menu.entryconfig("Delete Current Project", state="disabled")
        self._file_menu.entryconfig("Save Project As", state="disabled")
        self._file_menu.entryconfig("New Empty Project", state="disabled")
        self._file_menu.entryconfig("Load From Explorer", state="disabled")

        # Disables the file table close buttons
        self._files_table.enable_buttons(False)

    def _enable_buttons(self) -> None:
        """Re-enables the buttons and menu entries once processing is over."""

        self.log("Enabling the buttons")

        # Re-enabling buttons
        self._load_button['state'] = "enabled"
        self._select_all_button['state'] = 'normal'
        self._delete_all_button['state'] = 'enabled'
        self.set_unsaved_status()
        self._set_indicators()

        if self._current_project is not None:
            self._file_menu.entryconfig("Delete Current Project",
                                        state='normal')
        self._file_menu.entryconfig("Save Project As", state="normal")
        self._file_menu.entryconfig("New Empty Project", state="normal")
        self._file_menu.entryconfig("Load From Explorer", state="normal")

        # Re-enabling the file table close buttons
        self._files_table.enable_buttons(True)

    def _stop_processing(self, force: bool = True) -> None:
        """Empties the queue of images waiting for processing, and waits for
        the already started computation to end.

        This method is called when the user clicks on the Stop Processing
        button of the GUI. It is also called regularly by the processing thread
        so that the GUI gets back to normal operation mode once the last image
        has been processed.

        Args:
            force: If False, means that this method is called from the
                processing thread and won't have any effect unless the last
                image was just computed. If True, it is called by the user and
                will have effect.
        """

        # If called from the thread and there are images left to process, abort
        if not force and not self._queue.empty():
            return

        # If called from the thread but GUI already back to normal, abort
        if not force and self._stop_event.is_set():
            return

        # Sends a stop signal to the processing thread
        self.log("End of processing requested, setting the stop event")
        self._stop_event.set()

        # Wait for the job queue to be empty and then resets the GUI
        if force:
            self._processing_label['text'] = "Stopping"
            self._process_images_button["state"] = 'disabled'
            i = 1
            while not self._queue.empty():
                self.log("Processing queue not empty, waiting for the thread "
                         "to stop processing")
                self.update()
                # Indicating the user how many processes are left
                self._processing_label[
                    'text'] = f"Stopping {'.' * (i % 4)}{' ' * (4 - i % 4)}" \
                              f"(waiting for the current computation to " \
                              f"finish)"
                i += 1
                sleep(0.5)

        # Cleaning up the buttons and variables related to processing
        self.log("Cleaning up the processing-related objects")
        self._processed_images_count.set(0)
        self._img_to_process_count = 0
        self._process_images_button["state"] = 'enabled'
        self._process_images_button['text'] = "Process Images"
        self._process_images_button.configure(command=self._process_images)
        self._processing_label['text'] = ""

        # Putting the interface back into normal operation state
        self._enable_buttons()
        self.set_unsaved_status()

    def _process_images(self) -> None:
        """Prepares the interface and then sends to images to process to the
        processing thread."""

        self.log("User requested the images to be processed")

        # Getting the names of the images to process
        file_names = self._files_table.checked_files

        # If all checkboxes are unchecked, do nothing
        if not file_names:
            self.log("No images to process, aborting")
            return

        self.log(f"Processing the images: {', '.join(map(str, file_names))}")

        # Setting the buttons and variables related to processing
        self.log("Setting up the processing-related objects")
        self._processed_images_count.set(0)
        self._img_to_process_count = len(file_names)
        self._process_images_button['text'] = 'Stop Processing'
        self._process_images_button.configure(command=self._stop_processing)
        self._processing_label['text'] = f"0 of {len(file_names)} Images " \
                                         f"Processed "

        # Disabling most of the buttons and menu entries
        self._disable_buttons()
        self.update()

        # Sending the jobs to the processing thread
        self._stop_event.clear()
        self.log("Sending the jobs to the processing thread")
        for file in file_names:
            self._queue.put_nowait(
                (file,
                 self.settings.nuclei_colour.get(),
                 self.settings.fiber_colour.get(),
                 self.settings.fiber_threshold.get(),
                 self.settings.small_objects_threshold.get()))

    def _process_thread(self) -> None:
        """Main loop of the thread in charge of processing the images.

        Discards all the received jobs when the stop event is set.
        Otherwise, handles sequentially all the received jobs until the job
        queue is exhausted or until told to stop.
        Also periodically calls stop_processing so that the interface is rest
        after the last job in the queue was handled.
        """

        self.log("Starting the processing thread")

        # Looping forever until the stop_thread flag is raised
        while not self._stop_thread:

            # If the stop_event is set, we just have to discard all the jobs
            if self._stop_event.is_set():
                while not self._queue.empty():
                    self.log("Emptying the processing queue after the stop "
                             "event was set")
                    try:
                        self._queue.get_nowait()
                    except Empty:
                        pass

                sleep(1)

            # Case when there are jobs to handle
            elif not self._queue.empty():
                # Acquiring the next job in the queue
                try:
                    job = self._queue.get_nowait()
                    path, nuclei_color, fiber_color, fiber_threshold, \
                        small_objects_threshold = job
                    self.log(f"Processing thread received job: "
                             f"{', '.join(map(str, job))}")
                except Empty:
                    sleep(1)
                    continue

                # Not processing if the user wants to stop the computation
                if self._stop_event.is_set():
                    self.log("Stop event was set, aborting processing")
                    continue

                try:
                    # Now processing the image
                    file, nuclei_out, nuclei_in, fiber_contours, area = \
                        self._segmentation(path,
                                           nuclei_color,
                                           fiber_color,
                                           fiber_threshold,
                                           small_objects_threshold)
                    self.log(f"Segmentation returned file: {file}, "
                             f"nuclei out: {len(nuclei_out)}, "
                             f"nuclei in: {len(nuclei_in)}, "
                             f"fiber contours: {len(fiber_contours)}, "
                             f"area: {area}")

                    # Not updating if the user wants to stop the computation
                    if self._stop_event.is_set():
                        continue

                    # Updating the image data with the result of the processing
                    self.log("Passing precessed data to the files table")
                    self._files_table.input_processed_data(nuclei_out,
                                                           nuclei_in,
                                                           fiber_contours,
                                                           area, file)

                # Displaying any error in an error window
                except (Exception,) as exc:
                    messagebox.showerror("Error !",
                                         f"An error occurred while processing "
                                         f"the images !\nError : {exc}")
                    self._logger.exception("Exception caught wile processing "
                                           "image", exc_info=exc)

                # Updating the processed images count
                finally:
                    self._processed_images_count.set(
                        self._processed_images_count.get() + 1)

            # To lazily get the GUI back to normal once all the jobs are
            # completed
            else:
                self.log("Processing queue empty, requesting to stop the "
                         "processing")
                self._stop_processing(force=False)
                sleep(1)

        self.log("Processing thread finished")

    def _update_processed_images(self, _, __, ___) -> None:
        """Updates the display of the processed images count."""

        processed = self._processed_images_count.get()
        if processed:
            self._processing_label['text'] = f"{processed} of " \
                                             f"{self._img_to_process_count} " \
                                             f"Images Processed"

    def _select_images(self) -> None:
        """Opens a dialog window for selecting the images to load, then adds
        them in the files table."""

        self.log("Load button clicked by the user")

        # Getting the file names
        file_names = filedialog.askopenfilenames(
            filetypes=[('Image Files', ('.tif', '.png', '.jpg', '.jpeg',
                                        '.bmp', '.hdr'))],
            parent=self, title='Please select the images to open')

        if file_names:
            file_names = [Path(path) for path in file_names]
            self.log(f"User requested to load the images: "
                     f"{', '.join(map(str, file_names))}")

            # Enables the process images button
            self._process_images_button["state"] = 'enabled'

            # Adds the images to the table
            self._files_table.add_images(file_names)
            self.set_unsaved_status()

        else:
            self.log("No files selected by the user, aborting")

    def _set_image_channels(self) -> None:
        """Redraws the current image with the selected channels."""

        self.log("Setting the image channels to display")
        self._image_canvas.show_image()

    def _set_indicators(self) -> None:
        """Updates the display of fibers and nuclei according to the user
        selection."""
        
        # Updating the display
        self.log("Setting the image indicators to display")
        self._image_canvas.set_indicators()

    def _nuclei_colour_sel(self, _, __, ___) -> None:
        """Ensures there's no conflict in the chosen image channels for fibers
        and nuclei."""

        # Sets one of the channels back to its previous value in case of
        # conflict
        if self.settings.nuclei_colour.get() == \
                self.settings.fiber_colour.get():
            if self._previous_nuclei_colour.get() != \
                    self.settings.nuclei_colour.get():
                self.settings.fiber_colour.set(
                    self._previous_nuclei_colour.get())
            elif self._previous_fiber_colour.get() != \
                    self.settings.fiber_colour.get():
                self.settings.nuclei_colour.set(
                    self._previous_fiber_colour.get())

        # Sets the previous values variables
        self._previous_nuclei_colour.set(self.settings.nuclei_colour.get())
        self._previous_fiber_colour.set(self.settings.fiber_colour.get())

        # Finally, save and redraw the nuclei and fibers
        self._image_canvas.set_indicators()

    def _invert_checkboxes(self) -> None:
        """Called when clicking on the master checkbox.

        If all the boxes are checked, unchecks them all, otherwise checks them
        all.
        """

        self.log("Inverting all the checkboxes")

        if self._files_table.all_checked:
            for item in self._files_table.table_items:
                item.graph_elt.button_var.set(False)
        else:
            for item in self._files_table.table_items:
                item.graph_elt.button_var.set(True)

    def _delete_many_files(self) -> None:
        """Called when hitting the master delete button.

        Deletes all the files in the files table whose checkboxes are checked.
        """

        self.log("Delete all button clicked by the user")

        # Getting the list of files to delete, and checking it's not empty
        to_delete = self._files_table.checked_files
        if not to_delete:
            self.log("No file to delete, aborting")
            return

        # Actually deleting the files
        self.log(f"Deleting the files: {', '.join(map(str, to_delete))}")
        self._files_table.delete_image(tuple(to_delete))

    def _open_github(self) -> None:
        """Opens the project repository in a browser."""

        self.log("Opening the online documentation")
        open_new("https://github.com/TissueEngineeringLab/MyoFInDer")



# coding: utf-8

from tkinter import Toplevel, ttk, Scale
from screeninfo import get_monitors


class Settings_window(Toplevel):
    """Popup window for setting the parameters of the software."""

    def __init__(self, main_window) -> None:
        """Creates the window, arranges its layout and displays it.

        Args:
            main_window: The main window of the project.
        """

        super().__init__(main_window)

        # Setting the variables
        self._main_window = main_window
        self._settings = self._main_window.settings

        # Setting window properties
        self.resizable(False, False)
        self.grab_set()
        self.title("Settings")

        # Finish setting up the window
        self._set_layout()
        self.update()
        self._center()

    def destroy(self) -> None:
        """Before exiting, saves the settings to a file in the application
        folder."""

        if self._main_window.app_folder is not None:
            self._main_window.save_settings(self._main_window.app_folder)
        super().destroy()

    def _set_layout(self) -> None:
        """Creates the buttons and sliders, places them and displays them."""

        # General layout
        frame = ttk.Frame(self, padding="20 20 20 20")
        frame.grid(sticky='NESW')

        # Buttons for selecting the nuclei channel
        ttk.Label(frame, text="Nuclei Channel :").grid(column=0, row=0,
                                                       sticky='NE',
                                                       padx=(0, 10))
        nuclei_colour_r1 = ttk.Radiobutton(
            frame, text="Blue", variable=self._settings.nuclei_colour,
            value="blue")
        nuclei_colour_r1.grid(column=1, row=0, sticky='NW')
        nuclei_colour_r2 = ttk.Radiobutton(
            frame, text="Green", variable=self._settings.nuclei_colour,
            value="green")
        nuclei_colour_r2.grid(column=1, row=1, sticky='NW')
        nuclei_colour_r3 = ttk.Radiobutton(
            frame, text="Red", variable=self._settings.nuclei_colour,
            value="red")
        nuclei_colour_r3.grid(column=1, row=2, sticky='NW')

        # Buttons for selecting the fibers channel
        ttk.Label(frame, text="Fiber Channel :").grid(
            column=0, row=3, sticky='NE', pady=(10, 0), padx=(0, 10))
        fiber_colour_r1 = ttk.Radiobutton(
            frame, text="Blue", variable=self._settings.fiber_colour,
            value="blue")
        fiber_colour_r1.grid(column=1, row=3, sticky='NW', pady=(10, 0))
        fiber_colour_r2 = ttk.Radiobutton(
            frame, text="Green", variable=self._settings.fiber_colour,
            value="green")
        fiber_colour_r2.grid(column=1, row=4, sticky='NW')
        fiber_colour_r3 = ttk.Radiobutton(
            frame, text="Red", variable=self._settings.fiber_colour,
            value="red")
        fiber_colour_r3.grid(column=1, row=5, sticky='NW')

        # Buttons to chose whether to save the overlay images or not
        ttk.Label(frame, text='Save Images with Overlay :').grid(
            column=0, row=11, sticky='NE', pady=(10, 0), padx=(0, 10))
        ttk.Radiobutton(
            frame, text="On", variable=self._settings.save_overlay,
            value=1).grid(column=1, row=11, sticky='NW', pady=(10, 0))
        ttk.Radiobutton(
            frame, text="Off", variable=self._settings.save_overlay,
            value=0).grid(column=1, row=12, sticky='NW')

        # Slider to adjust the threshold for fiber detection
        ttk.Label(frame, text='Fiber detection threshold :').grid(
            column=0, row=13, sticky='E', pady=(10, 0), padx=(0, 10))

        fiber_threshold_slider_frame = ttk.Frame(frame)

        ttk.Label(fiber_threshold_slider_frame,
                  textvariable=self._settings.fiber_threshold,
                  width=3). \
            pack(side='left', anchor='w', fill='none', expand=False,
                 padx=(0, 20))

        Scale(fiber_threshold_slider_frame, from_=0, to=100,
              orient="horizontal",
              variable=self._settings.fiber_threshold, showvalue=False,
              length=150, tickinterval=20). \
            pack(side='left', anchor='w', fill='none', expand=False)

        fiber_threshold_slider_frame.grid(column=1, row=13, sticky='NW',
                                          pady=(10, 0))

        # Slider to adjust the small objects threshold
        ttk.Label(frame, text='Small Objects Threshold :').grid(
            column=0, row=14, sticky='E', pady=(10, 0), padx=(0, 10))

        threshold_slider_frame = ttk.Frame(frame)

        ttk.Label(threshold_slider_frame,
                  textvariable=self._settings.small_objects_threshold,
                  width=3). \
            pack(side='left', anchor='w', fill='none', expand=False,
                 padx=(0, 20))

        Scale(threshold_slider_frame, from_=0, to=100,
              variable=self._settings.small_objects_threshold,
              orient="horizontal", length=150, showvalue=False,
              tickinterval=25). \
            pack(side='left', anchor='w', fill='none', expand=False)

        threshold_slider_frame.grid(column=1, row=14, sticky='NW',
                                    pady=(10, 0))

    def _center(self) -> None:
        """Centers the popup window on the currently used monitor."""

        # Getting the current monitor
        monitors = get_monitors()
        candidates = [monitor for monitor in monitors if
                      0 <= self.winfo_x() - monitor.x <= monitor.width and
                      0 <= self.winfo_y() - monitor.y <= monitor.height]
        current_monitor = candidates[0] if candidates else monitors[0]

        # Getting the parameters of interest
        scr_width = current_monitor.width
        scr_height = current_monitor.height
        x_offset = current_monitor.x
        y_offset = current_monitor.y
        height = self.winfo_height()
        width = self.winfo_width()

        # Actually centering the window
        self.geometry('+%d+%d' % (x_offset + (scr_width - width) / 2,
                                  y_offset + (scr_height - height) / 2))



# coding: utf-8

from dataclasses import dataclass, field
from typing import Optional, List, Any, Tuple, Iterator, Union, Callable, Dict
from tkinter.ttk import Button, Separator, Label
from functools import partial
from tkinter import StringVar, IntVar, BooleanVar, Checkbutton, PhotoImage, \
    Frame, Event, EventType
from pathlib import Path
from platform import system


@dataclass
class Nucleus:
    """Class holding the data associated with a single nucleus."""

    x_pos: float
    y_pos: float
    tk_obj: Optional[int]
    color: str

    def __eq__(self, other: Any) -> bool:
        """Two nuclei are considered equal if and only if their x and y
        positions are equal."""

        if not isinstance(other, Nucleus):
            raise NotImplemented("Only two nuclei can be compared together")
        return self.x_pos == other.x_pos and self.y_pos == other.y_pos


@dataclass
class Fiber:
    """Class holding the data associated with a single fiber."""

    polygon: Optional[int]
    position: List[Tuple[float, float]] = field(default_factory=list)


@dataclass
class Nuclei:
    """Class for managing the data of all the nuclei in one image."""

    nuclei: List[Nucleus] = field(default_factory=list)

    def append(self, nuc: Nucleus) -> None:
        """Adds a nucleus to the list of nuclei."""

        self.nuclei.append(nuc)

    def remove(self, nuc: Nucleus) -> None:
        """Removes a given nucleus from the list of nuclei."""

        try:
            self.nuclei.remove(nuc)
        except ValueError:
            raise ValueError("No matching nucleus to delete")

    def reset(self) -> None:
        """Deletes all the saved Nucleus objects."""

        self.nuclei = list()

    def __iter__(self) -> Iterator[Nucleus]:
        """Returns an iterator over the saved Nucleus objects."""

        return iter(self.nuclei)

    @property
    def nuclei_in_count(self) -> int:
        """Returns the number of nuclei inside fibers."""

        return len([nuc for nuc in self.nuclei if nuc.color == 'in'])

    @property
    def nuclei_out_count(self) -> int:
        """Returns the number of nuclei outside fibers."""

        return len([nuc for nuc in self.nuclei if nuc.color == 'out'])

    def __len__(self) -> int:
        """Returns the number of nuclei."""

        return len(self.nuclei)


@dataclass
class Fibers:
    """Class for managing the data of all the fibers in one image."""

    area: float = field(default=0)
    fibers: List[Fiber] = field(default_factory=list)

    def append(self, fib: Fiber) -> None:
        """Adds a fiber to the list of fibers."""

        self.fibers.append(fib)

    def remove(self, fib: Fiber) -> None:
        """Removes a given fiber from the list of fibers."""

        try:
            self.fibers.remove(fib)
        except ValueError:
            raise ValueError("No matching fiber to delete")

    def reset(self) -> None:
        """Deletes all the saved Fiber objects."""

        self.fibers = list()
        self.area = 0

    def __iter__(self) -> Iterator[Fiber]:
        """Returns an iterator over the stored Fiber objects."""

        return iter(self.fibers)

    def __len__(self) -> int:
        """Returns the number of fibers."""

        return len(self.fibers)


class Graphical_element(Frame):
    """Class holding all the canvas elements for one image."""

    def __init__(self,
                 canvas: Frame,
                 number: int,
                 name: str,
                 delete_cmd: Callable,
                 check_cmd: Callable,
                 scroll_cmd: Callable,
                 select_cmd: Callable) -> None:
        """Sets the args, the attributes, the layout and the bindings.

        Args:
            canvas: The parent canvas in which the Frame is contained.
            number: The index of the Frame in the list of opened files.
            name: The name of the associated file, as displayed in the
                interface.
            delete_cmd: The callback command associated with the delete button.
            check_cmd: The callback command associated with the checkbox check.
            scroll_cmd: The callback command to call when a scroll action is
                performed.
            select_cmd: The callback command to call when a mouse click action
                is performed.
        """

        super().__init__(canvas)

        # Saving some attributes for later
        self._scroll_cmd = scroll_cmd
        self._select_cmd = select_cmd
        self._number = number

        # Instantiating variables
        self.selected = BooleanVar(self, value=False)
        self.button_var = BooleanVar(self, value=True)

        # Displaying the layout
        self._set_layout(number=number,
                         name=name,
                         delete_cmd=delete_cmd,
                         check_cmd=check_cmd)
        self.config(highlightthickness=3)
        self.config(highlightbackground='grey')

        # Adding traces and bindings
        self.selected.trace_add('write', self._select_trace)
        self.bind('<Enter>', self._hover)
        self.bind('<Leave>', self._hover)
        self._bind_wheel(self)
        self._bind_button(self)

    def _select_trace(self, _: str, __: str, ___: str) -> None:
        """A simple wrapper for calling the _hover method when the value of the
        selected boolean var is modified."""

        e = Event()
        e.type = None
        self._hover(e)

    def _bind_wheel(self, obj) -> None:
        """Method binding the mouse wheel action of an object to the
        associated scroll callback."""

        # Different wheel management in Windows and Linux
        if system() == "Linux":
            obj.bind('<4>', self._scroll_cmd)
            obj.bind('<5>', self._scroll_cmd)
        else:
            obj.bind('<MouseWheel>', self._scroll_cmd)

    def _bind_button(self, obj) -> None:
        """Method binding the mouse click action of an object to the
        associated click callback."""

        obj.bind('<ButtonPress-1>', partial(self._select_cmd,
                                            index=self._number))

    def _set_layout(self,
                    number: int,
                    name: str,
                    delete_cmd: Callable,
                    check_cmd: Callable) -> None:
        """Sets the layout of the Frame.

        Args:
            number: The index of the Frame in the list of opened files.
            name: The name of the associated file, as displayed in the
                interface.
            delete_cmd: The callback command associated with the delete button.
            check_cmd: The callback command associated with the checkbox check.
        """

        self.pack(expand=True, fill='x', pady=0, padx=0,
                  side='top', anchor='n')

        # The upper half of the frame
        self.frame_up = Frame(self)
        self.frame_up.pack(expand=True, fill='x', side='top', anchor='n')
        self._bind_wheel(self.frame_up)
        self._bind_button(self.frame_up)

        # Horizontal separator between the two halves of the frame
        self.sep_h = Separator(self, orient='horizontal')
        self.sep_h.pack(expand=True, fill='x', side='top', anchor='n')
        self._bind_wheel(self.sep_h)
        self._bind_button(self.sep_h)

        # The upper half of the frame
        self.frame_down = Frame(self)
        self.frame_down.pack(expand=True, fill='x', side='top', anchor='n')
        self._bind_wheel(self.frame_down)
        self._bind_button(self.frame_down)

        # The number of the frame in the list of all the frames
        self.label = Label(self.frame_up, text=str(number))
        self.label.pack(expand=False, fill='none', pady=15, padx=15,
                        side='left', anchor='w')
        self._bind_wheel(self.label)
        self._bind_button(self.label)

        # The button for deleting the frame
        self.close_button = Button(self.frame_up, text='X', width=2,
                                   command=delete_cmd)
        self.close_button.pack(expand=True, fill='none', side='right',
                               anchor='e', padx=(0, 2))
        self._bind_wheel(self.close_button)

        # The check button associated with the frame
        self.img = PhotoImage(width=1, height=1)
        self.check_button = Checkbutton(self.frame_up, image=self.img, width=6,
                                        height=24, variable=self.button_var,
                                        command=check_cmd)
        self.check_button.pack(expand=True, fill='none', side='right',
                               anchor='e')
        self._bind_wheel(self.check_button)

        # A vertical separator between the index of the frame and its name
        self.sep_v = Separator(self.frame_up, orient='vertical')
        self.sep_v.pack(expand=True, fill='y', side='left', anchor='w')
        self._bind_wheel(self.sep_v)
        self._bind_button(self.sep_v)

        # The name of the file associated with the frame
        self.name = Label(self.frame_up, text=name, width=40,
                          font=('Helvetica', 10))
        self.name.pack(expand=False, fill='none', padx=3, pady=15,
                       side='left', anchor='w')
        self._bind_wheel(self.name)
        self._bind_button(self.name)

        # The total number of detected nuclei
        self.total = Label(self.frame_down, text='error',
                           font=('Helvetica', 10))
        self.total.pack(expand=True, fill='x', side='left', anchor='w',
                        padx=0, pady=15)
        self._bind_wheel(self.total)
        self._bind_button(self.total)

        # The number of positive detected nuclei
        self.positive = Label(self.frame_down, text='error',
                              font=('Helvetica', 10))
        self.positive.pack(expand=True, fill='x', side='left', anchor='w',
                           padx=0, pady=15)
        self._bind_wheel(self.positive)
        self._bind_button(self.positive)

        # The ratio of positive nuclei over the total detected nuclei
        self.ratio = Label(self.frame_down, text='error',
                           font=('Helvetica', 10))
        self.ratio.pack(expand=True, fill='x', side='left', anchor='w',
                        padx=0, pady=15)
        self._bind_wheel(self.ratio)
        self._bind_button(self.ratio)

        # The percentage of area covered by fibers over the total image area
        self.area = Label(self.frame_down, text='error',
                          font=('Helvetica', 10))
        self.area.pack(expand=True, fill='x', side='left', anchor='w',
                       padx=0, pady=15)
        self._bind_wheel(self.area)
        self._bind_button(self.area)

    def _hover(self, event: Event) -> None:
        """Callback method called when the mouse enters or exits the Frame, or
        when the selected status of the Frame is updated."""

        # Highlight the frame if it's selected or being hovered
        if self.selected.get() or event.type == EventType.Enter:
            color = 'black'
        else:
            color = 'grey'

        self.config(highlightbackground=color)


@dataclass
class Table_entry:
    """Class holding all the information associated with one image."""

    path: Union[Path, str]
    nuclei: Nuclei
    fibers: Fibers
    graph_elt: Optional[Graphical_element] = None

    @property
    def save_version(self):
        """Returns a simpler version of the current class, with no canvas
        object and with the path truncated to only the file name."""

        return Table_entry(path=self.path.name,
                           nuclei=self.nuclei,
                           fibers=self.fibers,
                           graph_elt=None)


@dataclass
class Table_items:
    """Class storing all the Table_entry classes of the opened project.

    It also implements many helper functions for simplifying the code in the
    rest of the project.
    """

    entries: List[Table_entry] = field(default_factory=list)
    current_index: Optional[int] = None

    def __getitem__(self, item: Path) -> Table_entry:
        """Returns the Table_entry object whose path corresponds to the given
        one."""

        # Searches for the first (hopefully the only) entry with a given path
        for entry in self.entries:
            if entry.path == item:
                return entry

        # In the (unlikely) case when there's no entry for the requested path
        raise KeyError

    def __iter__(self) -> Iterator[Table_entry]:
        """Returns an iterator over the stored Table_entry objects."""

        return iter(self.entries)

    def __bool__(self) -> bool:
        """Returns True if there's at least one stored Table_entry, False
        otherwise."""

        return bool(self.entries)

    def __len__(self) -> int:
        """Returns the number of stored Table_entry objects."""

        return len(self.entries)

    @property
    def file_names(self) -> List[Path]:
        """Returns a list of all the paths of the stored Table_entry
        objects."""

        return [entry.path for entry in self]

    @property
    def current_entry(self) -> Optional[Table_entry]:
        """Returns the Table_entry instance corresponding to the currently
        displayed image."""

        if self.selected is None:
            return

        return self[self.selected]

    @property
    def selected(self) -> Optional[Path]:
        """Returns the path of the Table_entry instance corresponding to the
        currently displayed image."""

        for entry in self.entries:
            if entry.graph_elt.selected.get():
                return entry.path

    @property
    def save_version(self):
        """Returns a simpler version of the current class, with no canvas
        object and with the paths of the Table_entry objects truncated to only
        the file name."""

        return Table_items(entries=[entry.save_version for entry
                                    in self.entries])

    def reset(self) -> None:
        """Resets all the graphics and deletes all the stored Table_entry
        objects."""

        self.reset_graphics()
        self.entries = list()

    def reset_graphics(self) -> None:
        """Deletes all the displayed Frames of the Table_entries, but keeps the
        other attributes of the entries."""

        for entry in self:
            entry.graph_elt.pack_forget()
            entry.graph_elt.destroy()
            del entry.graph_elt

    def append(self, entry: Table_entry) -> None:
        """Adds a Table_entry object to the list of the stored ones."""

        self.entries.append(entry)

    def remove(self, path: Path) -> None:
        """Removes the Table_entry object corresponding to the given path from
        the list of the stored ones."""

        try:
            self.entries.remove(self[path])
        except KeyError:
            raise ValueError(f"No table entry associated with the path {path}")

    def index(self, path: Path) -> int:
        """Returns the position of the Table_entry corresponding to the given
        path in the list of all the stored Table_entry objects."""

        for i, entry in enumerate(self):
            if entry.path == path:
                return i

        raise ValueError(f"No table entry associated with the path {path}")


@dataclass
class Selection_box:
    """Class holding the data associated with the selection box of the image
    canvas."""

    x_start: Optional[int] = None
    y_start: Optional[int] = None
    x_end: Optional[int] = None
    y_end: Optional[int] = None
    tk_obj: Optional[int] = None

    def __bool__(self) -> bool:
        """Returns True if all four corners of the selection box have been
        defined, else False."""

        return all(item is not None for item in (self.x_start, self.y_start,
                                                 self.x_end, self.y_end))

    def is_inside(self, x: float, y: float) -> bool:
        """Checks whether a given point is inside or outside the selection box.

        Args:
            x: The x coordinate of the point to check.
            y: The y coordinate of the point to check.

        Returns:
            True if the given coordinates lie inside the selection box, False
            otherwise.
        """

        if not bool(self):
            return False

        x_min = min(self.x_start, self.x_end)
        x_max = max(self.x_start, self.x_end)
        y_min = min(self.y_start, self.y_end)
        y_max = max(self.y_start, self.y_end)

        return (x_min <= x <= x_max) and (y_min <= y <= y_max)

    @property
    def started(self) -> bool:
        """Returns True if the first corner of the selection box has been
        defined, else False."""

        return self.x_start is not None and self.y_start is not None

    @property
    def area(self) -> int:
        """Returns the area of the selection box."""

        return abs((self.x_end - self.x_start) * (self.y_end - self.y_start))


@dataclass
class Settings:
    """Class holding the current values of all the settings."""

    # Here a default factory is needed so that the tkinter vars are not
    # instantiated before the tkinter app is initialized
    fiber_colour: StringVar = field(
        default_factory=partial(StringVar, value="green", name='fiber_colour'))
    nuclei_colour: StringVar = field(
        default_factory=partial(StringVar, value="blue", name='nuclei_colour'))
    save_overlay: BooleanVar = field(
        default_factory=partial(BooleanVar, value=False, name='save_overlay'))
    fiber_threshold: IntVar = field(
        default_factory=partial(IntVar, value=25, name='fiber_threshold'))
    small_objects_threshold: IntVar = field(
        default_factory=partial(IntVar, value=20, name='small_objects'))
    blue_channel_bool: BooleanVar = field(
        default_factory=partial(BooleanVar, value=True, name='blue_channel'))
    green_channel_bool: BooleanVar = field(
        default_factory=partial(BooleanVar, value=True, name='green_channel'))
    red_channel_bool: BooleanVar = field(
        default_factory=partial(BooleanVar, value=False, name='red_channel'))
    show_nuclei: BooleanVar = field(
        default_factory=partial(BooleanVar, value=True, name='show_nuclei'))
    show_fibers: BooleanVar = field(
        default_factory=partial(BooleanVar, value=False, name='show_fibers'))

    def update(self, settings: Dict[str, Any]) -> None:
        """Updates the values of the settings based on the provided
        dictionary."""

        for key, value in settings.items():
            getattr(self, key).set(value)

    def __str__(self) -> str:
        """"""

        settings = (self.fiber_colour, self.nuclei_colour,
                    self.save_overlay, self.fiber_threshold,
                    self.small_objects_threshold, self.blue_channel_bool,
                    self.green_channel_bool, self.red_channel_bool,
                    self.show_nuclei, self.show_fibers)

        return ', '.join(f"{setting}: {setting.get()}" for setting in settings)



# coding: utf-8

"""Executable file for running the application"""

from . import Main_window
import logging
from sys import stdout
from pathlib import Path
import argparse
from platform import system


if __name__ == "__main__":

    # Parser for parsing the command line arguments of the script
    parser = argparse.ArgumentParser(
        description="Starts MyoFInDer, with optional flags for specifying "
                    "whether to record log messages and whether the module was"
                    " started by the application or independently.")
    parser.add_argument('-n', '--nolog', action='store_false',
                        help="If provided, the log messages won't be displayed"
                             " and recorded. Otherwise, they are by default.")
    args = parser.parse_args()

    # Retrieving the command-line argument
    log = args.nolog

    # Setting the path to the application folder
    if system() in ('Linux', 'Darwin'):
        app_folder = Path.home() / '.MyoFInDer'
    elif system() == 'Windows':
        app_folder = (Path.home() / 'AppData' / 'Local' / 'MyoFInDer')
    else:
        app_folder = None

    # Creating the application folder, if needed
    if app_folder is not None:
        try:
            app_folder.mkdir(parents=True, exist_ok=True)
        except FileNotFoundError:
            app_folder = None

    # Setting up the logger
    logger = logging.getLogger("MyoFInDer")
    logger.setLevel(logging.INFO)

    # Setting up the handlers only if logging is enabled
    if log:
        handler_console = logging.StreamHandler(stream=stdout)
        handler_console.setLevel(logging.INFO)

        # Setting up the formatter
        formatter = logging.Formatter('%(asctime)s %(name)-8s %(message)s')
        handler_console.setFormatter(formatter)

        logger.addHandler(handler_console)

        # Setting up the file handler only if a location is provided
        if app_folder is not None:
            handler_file = logging.FileHandler(app_folder / 'logs.txt',
                                               mode='w')
            handler_file.setLevel(logging.INFO)
            handler_file.setFormatter(formatter)
            logger.addHandler(handler_file)

    # Normal workflow
    try:
        logger.log(logging.INFO, "Launching MyoFInDer")
        window = Main_window(app_folder)
        window.mainloop()
        logger.log(logging.INFO, "MyoFInDer terminated gracefully")

    # Displaying the exception and waiting for the user to close the console
    except (Exception,) as exc:
        logger.exception("MyoFInDer encountered an error while running !",
                         exc_info=exc)
        # Only leave the console on if logging is enabled
        if log:
            print("\n\nPress ENTER to exit ...")
            input()



# coding: utf-8

"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from setuptools import find_namespace_packages, setup

# Reading version from __version__.py file
with open('src/myofinder/__version__.py') as file:
  for line in file:
    if line.startswith('__version__'):
      __version__ = line.split("'")[1]

setup(
  # Description of the project
  name='myofinder',
  version=__version__,
  description='Automatic calculation of the fusion index by AI segmentation',
  long_description="""Myoblast Fusion Index Determination
  Myofinder is is a Python module for the automatic calculation of the fusion 
  index on florescence-stained images of muscle cultures. It provides a simple 
  and user-friendly interface managing several images, processing all or part 
  of them, visualizing the output, and manually correcting it if needed. It 
  relies on the DeepCell library for performing the image segmentation.
  """,
  keywords='segmentation,fusion index,automation,muscle culture',
  license='GPL V3',
  classifiers=['Development Status :: 5 - Production/Stable ',
               'Intended Audience :: Science/Research',
               'License :: OSI Approved :: GNU General Public License v3 '
               '(GPLv3)',
               'Natural Language :: English',
               'Operating System :: OS Independent',
               'Programming Language :: Python :: 3.7',
               'Programming Language :: Python :: 3.8',
               'Programming Language :: Python :: 3.9',
               'Programming Language :: Python :: 3.10',
               'Topic :: Scientific/Engineering :: Artificial Intelligence',
               'Topic :: Scientific/Engineering :: Bio-Informatics',
               'Topic :: Scientific/Engineering :: Image Recognition'
               ],

  # URLs of the project
  url='https://github.com/TissueEngineeringLab/MyoFInDer',
  download_url='https://pypi.org/project/myofinder/#files',
  project_urls={
    'Documentation': 'https://tissueengineeringlab.github.io/MyoFInDer/',
    'Source': 'https://github.com/TissueEngineeringLab/MyoFInDer'},

  # Information on the author
  author='Antoine Weisrock',
  author_email='antoine.weisrock@kuleuven.be',
  maintainer='Antoine Weisrock',
  maintainer_email='antoine.weisrock@gmail.com',

  # Packaging information
  packages=find_namespace_packages(where="src", exclude=list()),
  package_dir={"": "src"},
  include_package_data=True,
  package_data={"myofinder": ["app_images/*"]},
  ext_package='myofinder',
  ext_modules=list(),

  # Installation requirements
  python_requires=">=3.7,<3.11",
  install_requires=['Pillow>=9.0.0', 'opencv-python>=4.5', 'DeepCell>=0.10.0',
                    'XlsxWriter>=3.0.0', 'screeninfo>=0.7', 'numpy<1.24'],
  extras_require=dict(),

  # Others
  zip_safe=False,
)



# coding: utf-8

from deepcell.applications import Mesmer
import numpy as np
import numpy.ma as ma
from pathlib import Path
import cv2
from typing import List, Tuple, Any

from .tools import check_image

# Table for converting color strings to channels, assuming BGR images
numpy_color_to_int = {"red": 0,
                      "green": 1,
                      "blue": 2}


class Image_segmentation:
    """Class for processing images, detecting fibers and nuclei."""

    def __init__(self) -> None:
        """Simply loads the Mesmer library."""

        self._app = Mesmer()

    def __call__(self,
                 path: Path,
                 nuclei_color: str,
                 fiber_color: str,
                 fiber_threshold: int,
                 small_objects_threshold: int) -> \
            (Path, List[Tuple[np.ndarray, np.ndarray]],
             List[Tuple[np.ndarray, np.ndarray]], Tuple[Any], float):
        """Computes the nuclei positions and optionally the fibers positions.

        Also returns whether the nuclei are inside or outside the fibers.

        Args:
            path: The path to the image to process.
            nuclei_color: The color of the nuclei, as a string.
            fiber_color: The color of the fibers, as a string.
            fiber_threshold: The gray level threshold above which a pixel is
                considered to be part of a fiber.
            small_objects_threshold: Objects whose area is lower than this
                value (in pixels) will not be considered.

        Returns:
            The list of nuclei outside the fibers, the list of nuclei inside
            the fibers, the list of fiber contours, and the ratio of fiber area
            over the total area.
        """

        # Converting colors from string to int
        colors = [numpy_color_to_int[nuclei_color],
                  numpy_color_to_int[fiber_color]]

        # Loads the image and keeps only the nuclei and fibers channels
        image = check_image(path)

        # The image couldn't be loaded
        if image is None:
            raise IOError("Could not load the image for segmentation, "
                          "aborting !")

        # Removing the scale bar
        image[(image[:, :, 0] > 50) &
              (image[:, :, 1] > 50) &
              (image[:, :, 2] > 50)] = (0, 0, 0)

        nuclei_channel = image[:, :, colors[0]]
        fiber_channel = image[:, :, colors[1]]

        del image

        # Default parameters
        radius = 10
        maxima_threshold = 0.1
        interior_threshold = 0.01
        maxima_smooth = 0
        interior_smooth = 0
        maxima_index = 0
        interior_index = -1
        label_erosion = 0
        fill_holes_threshold = 0
        pixel_expansion = None
        maxima_algorith = 'h_maxima'

        # Actual nuclei detection function
        labeled_image = self._app.predict(
            image=np.stack((nuclei_channel,
                            nuclei_channel), axis=-1)[np.newaxis, :],
            batch_size=1,
            image_mpp=None,
            pad_mode='constant',
            compartment='nuclear',
            preprocess_kwargs=dict(),
            postprocess_kwargs_whole_cell=dict(),
            postprocess_kwargs_nuclear={
                'radius': radius,
                'maxima_threshold': maxima_threshold,
                'interior_threshold': interior_threshold,
                'maxima_smooth': maxima_smooth,
                'interior_smooth': interior_smooth,
                'maxima_index': maxima_index,
                'interior_index': interior_index,
                'label_erosion': label_erosion,
                'small_objects_threshold': small_objects_threshold,
                'fill_holes_threshold': fill_holes_threshold,
                'pixel_expansion': pixel_expansion,
                'maxima_algorith': maxima_algorith,
            })

        # Removing useless axes on the output and nuclei channel
        labeled_image = np.squeeze(labeled_image)

        # Getting the fiber mask
        mask = self._get_fiber_mask(fiber_channel,
                                    nuclei_channel,
                                    fiber_threshold)

        del fiber_channel

        # Calculating the area of fibers over the total area
        area = np.count_nonzero(mask) / mask.shape[0] / mask.shape[1]

        # Finding the contours of the fibers
        mask_8_bits = (mask * 255).astype('uint8')
        fiber_contours, _ = cv2.findContours(mask_8_bits, cv2.RETR_LIST,
                                             cv2.CHAIN_APPROX_SIMPLE)
        fiber_contours = tuple(map(np.squeeze, fiber_contours))

        # Getting the position of the nuclei
        nuclei_out, nuclei_in = self._get_nuclei_positions(
            labeled_image, mask, nuclei_channel, 0.4)

        return path, nuclei_out, nuclei_in, fiber_contours, area

    @staticmethod
    def _get_fiber_mask(fiber_channel: np.ndarray,
                        nuclei_channel: np.ndarray,
                        threshold: int) -> np.ndarray:
        """Applies several images processing methods to the fiber channel of
        the image to smoothen the outline.

        Args:
            fiber_channel: The channel of the image containing the fibers.
            nuclei_channel: The channel of the image containing the nuclei.
            threshold: The gray level threshold above which a pixel is
                considered to be part of a fiber.

        Returns:
            A boolean mask containing the position of the fibers
        """

        # First, apply a base threshold
        kernel = np.ones((4, 4), np.uint8)
        _, processed = cv2.threshold(fiber_channel, threshold, 255,
                                     cv2.THRESH_BINARY)

        # Opening, to remove noise in the background
        processed = cv2.morphologyEx(processed,
                                     cv2.MORPH_OPEN, kernel)

        # Closing, to remove noise inside the fibers
        processed = cv2.morphologyEx(processed, cv2.MORPH_CLOSE, kernel)

        # Inverting black and white
        processed = cv2.bitwise_not(processed)

        # Find contours of the holes inside the fibers
        contours, _ = cv2.findContours(processed, cv2.RETR_TREE,
                                       cv2.CHAIN_APPROX_SIMPLE)

        for contour in contours:

            mask = np.zeros_like(processed)
            cv2.drawContours(mask, contour, -1, 255, -1)

            if np.average(nuclei_channel[mask == 255]) > 50:
                cv2.drawContours(processed, contour, -1, 0, -1)

        # Inverting black and white again
        processed = cv2.bitwise_not(processed)

        # Creating a boolean mask, later used to generate a masked array
        mask = processed.astype(bool)

        return mask

    @staticmethod
    def _get_nuclei_positions(labeled_image: np.ndarray,
                              mask: np.ndarray,
                              nuclei_channel: np.ndarray,
                              fiber_threshold: float = 0.85,
                              nuclei_threshold: int = 50) -> \
            (List[Tuple[np.ndarray, np.ndarray]],
             List[Tuple[np.ndarray, np.ndarray]]):
        """Computes the center of the nuclei and determines whether they're
        positive or not.

        Args:
            labeled_image: The image containing the nuclei.
            mask: The boolean mask of the fibers.
            nuclei_channel: The channel of the image containing the nuclei.
            fiber_threshold: Fraction of area above which a nucleus is
                considered to be inside a fiber.

        Returns:
            The list of the centers of nuclei outside of fibers, and the list
            of centers of nuclei inside fibers
        """

        nuclei_out_fiber = []
        nuclei_in_fiber = []

        # Masking the image containing the nuclei with the boolean mask of the
        # fibers
        masked_image = ma.masked_array(labeled_image, mask)

        # Each gray level on the nuclei image corresponds to one nucleus
        nb_nuc = np.max(labeled_image)

        # Iterating through the detected nuclei
        for i in range(1, nb_nuc + 1):

            # Getting a list of the detected pixels
            nucleus_y, nucleus_x = np.where(labeled_image == i)
            center_x, center_y = np.mean(nucleus_x), np.mean(nucleus_y)

            # Aborting for the current nucleus if it's not bright enough
            if np.average(nuclei_channel[nucleus_y,
                                         nucleus_x]) < nuclei_threshold:
                continue

            # Determining whether the nucleus is positive or negative
            in_fiber = ma.count_masked(masked_image[nucleus_y, nucleus_x])
            if in_fiber < fiber_threshold * nucleus_x.shape[0]:
                nuclei_out_fiber.append((center_x, center_y))
            else:
                nuclei_in_fiber.append((center_x, center_y))

        return nuclei_out_fiber, nuclei_in_fiber
